\name{pwlik_mev}
\alias{pwlik_mev}
\title{Penalized pairwise log likelihood for multivariate extreme value models}
\description{
Computes the penalized pairwise loglikelihood function for the multivariate extreme value model descrived in Padoan et al. 2010. 
}
\usage{
pwlik_mev(z, theta, sets, n, q, bias.reduction=FALSE)
}
\arguments{
  \item{z}{a matrix containing independent realizations from a max-stable random field.}
\item{theta}{the vector of model's parameters}
\item{n}{number of independent realizations from the max stable random field.} 
\item{q}{number of site locations at which the max-stable random field is observed.}
\item{bias.reduction}{Logical. Indicates whether bias reduction is going to be performed by adding a penalization to the pairwise loglikelihood.}
}
\details{To simulate data properly see function \code{\link{rmaxstable}} from package \code{\link{SpatialExtremes}} as well as the example below.}

\value{
	The function returns the value of the pairwise loglikleihood or its penalized version.}

\references{
Padoan, S. A., Ribatet, M., and Sisson, S. A. (2005). Likelihood-Based Inference for Max-Stable Processes. \emph{JASA}, \bold{105}, 263--277.

}
\author{Nicola Lunardon}

\examples{
#set parameters and simulate data
##generate sites
set.seed(1)
#setting as in Padoan et al 2010 - section 4
n.sites <- 50
locs <- cbind(runif(n.sites, 0, 40), runif(n.sites, 0, 40))
#consider all the pairs
d <- max(locs)+1
sets <- my_sets(locs, d, 1)

#entries of the covariance matrix
#diagonal elements
theta <- c(2e3,3e3)
#off diagonal element
theta <- c(theta, 1.5e3)

#number of replicates
n <- 50

if(require(SpatialExtremes)){
	data_sim <- rmaxstab(n, locs, "gauss", cov11 = theta[1], cov12 = theta[3], cov22 = theta[2], grid = FALSE)
}

if(require(alabama)){
	#function to put constraints on the entries of the covariance matrix
	#to be used in function auglag (pkg alabama)
	hin <- function(x, ...){
		h <- rep(NA, 3)
		h[1] <- x[1]-1e-6#sig11>0
		h[2] <- x[2]-1e-6#sig22>0
		h[3] <- x[1]*x[2]-x[3]^2-1e-6#determinant>0
		h
	}


	#standard maximum pairwise likelihood estimator
	obj.std <- auglag(par=theta, fn=function(y) -pwlik_mev(theta=y, z=data_sim, sets=sets, n=n, q=n.sites, bias.reduction=FALSE), 
						 hin=hin,
						 control=list(iter.max=500), 
						 control.outer=list(itmax=1e4, trace=FALSE, method="nlminb"))


	#bias-reduced maximum pairwise likelihood estimator
	obj.rbr <- auglag(par=theta, fn=function(y) -pwlik_mev(theta=y, z=data_sim, sets=sets, n=n, q=n.sites, bias.reduction=TRUE), 
						 hin=hin,
						 control=list(iter.max=500), 
						 control.outer=list(itmax=1e4, trace=FALSE, method="nlminb"))


}


}
\keyword{pairwise likelihood.}

