## Copyright (C) 2022 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!
##
## It is assumed that the user has a Julia and R installation with all
## packages required by the current script and any other script
## included.
##
##
## Code used by scripts that reproduce the results in Example 5.3

module auto_symmetric ## see Wolters (2017)

using LinearAlgebra
using Graphs
using Distributions
using Autologistic
using Optim
using FiniteDiff
using RCall

struct Data
    Z::Vector ## A vector with elements from {-1, 1}
    X::Matrix ## A matrix of covariates corresponding to each element of Z
    G::SimpleGraph{Int64} ## A graph setting the neighbours
end


function nobs(data::Data)
    ## add size checks here
    length(data.Z)
end

function nvar(data::Data)
    ## add size checks here
    size(data.X)[2]
end

function auxilliary(theta::Vector,
                    data::Data)
    p = nvar(data)
    n = nobs(data)
    beta = theta[1:p]
    lambda = theta[p + 1]
    y = @. (data.Z + 1) / 2 
    X = data.X
    z = data.Z
    nei = map(i -> neighbors(data.G, i), 1:n)
    dep = map(i -> sum(z[nei[i]]), 1:n)
    deta = 2 .* hcat(X, dep)
    probs = cdf.(Logistic(), deta * theta)
    Dict("p" => p,
         "n" => n,
         "y" => y,
         "nei" => nei,
         "beta" => beta,
         "probs" => probs,
         "deta" => deta)
end

function objective(theta::Vector,
                   data::Data,
                   br::Bool = false;
                   subsets::Union{Nothing, Vector{Vector{Int64}}} = nothing,
                   plugin::Bool = false)
    aux = auxilliary(theta, data)
    y = aux["y"]
    probs = aux["probs"]
    if br
        j = jmat(theta, data, aux = aux)
        if subsets isa Nothing ## use naive, we should not really want that
            e = emat_naive(theta, data, aux = aux)
        else
            e = emat_sub(theta, data, aux = aux, subsets = subsets)
        end
        penalty = try
            - tr(inv(j) * e) / 2
        catch
            NaN
        end
        if plugin
            penalty = try
                penalty + log(det(j)) / aux["n"]
            catch
                NaN
            end
        end
    else
        penalty = 0.0
        if plugin
            j = jmat(theta, data, aux = aux)
            penalty = try
                penalty + log(det(j)) / aux["n"]
            catch
                NaN
            end
        end
    end
    sum(@. y * log(probs) + (1 - y) * log(1 - probs)) + penalty
end

function grad(theta::Vector,
              data::Data,
              br::Bool = false;
              subsets::Union{Nothing, Vector{Vector{Int64}}} = nothing,
              plugin::Bool = false)
    aux = auxilliary(theta, data)
    y = aux["y"]
    if br | plugin
        obj = th -> objective(th, data, br, subsets = subsets, plugin = plugin) 
        FiniteDiff.finite_difference_gradient(obj, theta)
    else
        aux["deta"]' * (y - aux["probs"])
    end
end

function jmat(theta::Vector,
              data::Data;
              aux::Union{Nothing, Dict} = nothing)
    if aux isa Nothing
        aux = auxilliary(theta, data)
    end
    xtilde = aux["deta"]
    probs = aux["probs"] 
    v = @. probs * (1 - probs)
    (xtilde .* v)' * xtilde
end

function emat_naive(theta::Vector,
                    data::Data;
                    aux::Union{Nothing, Dict} = nothing)
    if aux isa Nothing
        aux = auxilliary(theta, data)
    end
    y = aux["y"]
    xtilde = aux["deta"]
    probs = aux["probs"] 
    gr_contr = aux["deta"] .* (y - aux["probs"])
    sum(map(i -> gr_contr[i, :] * gr_contr[i, :]', 1:aux["n"]))
end

function emat_sub(theta::Vector,
                  data::Data;
                  aux::Union{Nothing, Dict} = nothing,
                  subsets::Union{Nothing, Vector{Vector{Int64}}} = nothing)
    if aux isa Nothing
        aux = auxilliary(theta, data)
    end
    if subsets isa Nothing
        subsets = [collect(1:nobs(data))]
    end
    nsubsets = length(subsets)  
    y = aux["y"]
    xtilde = aux["deta"]
    probs = aux["probs"]
    xtilde = aux["deta"]
    resids = (y - aux["probs"])   
    sub_scores = map(s -> xtilde[s, :]' * resids[s], subsets)
    sum(map(r -> sub_scores[r] * sub_scores[r]', 1:nsubsets))
end

function vcov(theta::Vector,
              data::Data;
              aux::Union{Nothing, Dict} = nothing,
              subsets::Union{Nothing, Vector{Vector{Int64}}} = nothing)
    if aux isa Nothing
        aux = auxilliary(theta, data)
    end
    if subsets isa Nothing ## use naive, we should not really want that
        e = emat_naive(theta, data, aux = aux)
    else
        e = emat_sub(theta, data, aux = aux, subsets = subsets)
    end
    j = jmat(theta, data, aux = aux)
    out = try
        j_inv = inv(j)
        j_inv * (e * j_inv')
    catch
        fill(NaN, size(j))
    end
    out
end

function se(theta::Vector,
            data::Data;
            aux::Union{Nothing, Dict} = nothing,
            subsets::Union{Nothing, Vector{Vector{Int64}}} = nothing)
    vars = diag(vcov(theta, data, aux = aux, subsets = subsets))
    vars[vars .< 0] .= NaN
    sqrt.(vars)
end

function zstats(estimates::Vector,
                data::Data,
                ses::Union{Vector, Nothing} = nothing;
                subsets::Union{Nothing, Vector{Vector{Int64}}} = nothing)
    if ses isa Nothing
        ses = se(estimates, data, subsets = subsets)
    end
    zstats = estimates  ./ ses
    pvalues = 2 * cdf(Normal(0, 1), -abs.(zstats))
    Dict("estimates" => estimates,
         "ses" => ses,
         "z" => zstats,
         "pvalue" => pvalues)
end

function fit(theta::Vector,
             data::Data,
             br::Bool = false;
             subsets::Union{Nothing, Vector{Vector{Int64}}} = nothing,
             Mstart::Bool = true, ## Shall we start BR from the M estimates
             plugin::Bool = false,
             optim_method = LBFGS(),
             optim_options = Optim.Options(),
             auto_diff = true,
             optim_arguments...)
    objM = th -> -objective(th, data, false, plugin = plugin)
    if auto_diff
        objM = OnceDifferentiable(objM, theta; autodiff = :forward)
    end
    if br
        objRBM = th -> -objective(th, data, true, plugin = plugin, subsets = subsets)
        if auto_diff
            objRBM = OnceDifferentiable(objRBM, theta; autodiff = :forward)
        end
        if Mstart
            theta = optimize(objM,
                             theta,
                             optim_method,
                             optim_options;
                             optim_arguments...).minimizer
        end
        optimize(objRBM,
                 theta,
                 optim_method,
                 optim_options;
                 optim_arguments...)
    else
        optimize(objM,
                 theta,
                 optim_method,
                 optim_options;
                 optim_arguments...)
    end        
end                

function simulate(theta::Vector,
                  data::Data,
                  R::Int64)
    p = nvar(data)
    beta = theta[1:p]
    lambda = theta[p + 1]
    model = ALRsimple(data.G, data.X, Y = data.Z, coding = (-1, 1), centering = none,
                      β = beta,
                      λ = lambda)
    Autologistic.sample(model, R,
                        method = Gibbs,#perfect_bounding_chain,
                        average = false);
end

function is_separated(data::Data)
    pall = nvar(data) + 1
    Xtilde = auto_symmetric.auxilliary(zeros(pall), data)["deta"]
    ZZ = @. (data.Z + 1) / 2
    rcopy(R"library(detectseparation)
            mod_sep <- glm($ZZ ~ -1 + $Xtilde, family = binomial('logit'), method = 'detect_separation')
            is.infinite(coef(mod_sep))")
end

function boot(theta::Vector,
              data::Data,
              br::Bool = false,
              R::Int64 = 1000;
              subsets::Union{Nothing, Vector{Vector{Int64}}} = nothing,
              plugin::Bool = false,
              verbose::Int64 = 100,
              Mstart::Bool = false,
              optim_method = LBFGS(),
              optim_options = Optim.Options(),
              alt_start::Union{Nothing, Vector{Float64}} = nothing,
              optim_arguments...)
    p = nvar(data)    
    s_responses = simulate(theta, data, R)
    coefs = Matrix{Float64}(undef, R, p + 1)
    is_infinite = Matrix{Bool}(undef, R, p + 1)
    converged_optim = Vector{Bool}(undef, R)
    converged = Matrix{Bool}(undef, R, p + 1)
    objective_values = Vector{Float64}(undef, R)
    for r in 1:R
        cdat = Data(s_responses[:, r], data.X, data.G)
        is_infinite[r,:] = is_separated(cdat)
        progress(r, R, "Separation detection ", verbose)
    end
    Threads.@threads for r in 1:R        
        try
            if alt_start isa Nothing
                start = theta
            else
                start = alt_start
            end
            cdat = Data(s_responses[:, r], data.X, data.G)
            result = fit(start,
                         cdat,
                         br,
                         subsets = subsets,
                         plugin = plugin,
                         Mstart = Mstart,
                         optim_method = optim_method,
                         optim_options = optim_options,
                         optim_arguments...)
            coefs[r, :] = Optim.minimizer(result)
            objective_values[r] = -Optim.minimum(result)
            converged_optim[r] = Optim.converged(result)
            if plugin
                converged[r, :] .= converged_optim[r]
            else
                converged[r, :] = converged_optim[r] .& .!is_infinite[r,:]
            end
        catch
            coefs[r, :] = fill(NaN, p + 1)
            objective_values[r] = NaN
            converged_optim[r] = false
            converged[r, :] .= false
        end
        progress(r, R, "Bootstrap ", verbose)
    end
    Dict("responses" => s_responses,
         "theta" => theta,
         "coefficients" => coefs,
         "objective_values" => objective_values,
         "converged_optim" => converged_optim,
         "converged" => converged,
         "is_infinite" => is_infinite)
end

function dropna(x::Vector)
    z = copy(x)
    z[.!isnan.(z)]
end

function wrangle(v::Array{Float64, 2}, id::Matrix{Bool}, C::Float64)
    z = copy(v)
    if C < Inf
        z[.!id] .= NaN
        z[abs.(z) .> C] .= NaN
    end
    z
end

function summarize_boot(object::Dict; C::Float64 = Inf, alpha::Float64 = 0.05)
    coefs = wrangle(object["coefficients"], object["converged"], C)
    means = mapslices(col -> mean(dropna(col)), coefs, dims = 1)'
    sds = mapslices(col -> sqrt.(var(dropna(col))), coefs, dims = 1)'
    quants_upp = mapslices(col -> quantile(dropna(col), 1 - alpha/2), coefs, dims = 1)'
    quants_low = mapslices(col -> quantile(dropna(col), alpha/2), coefs, dims = 1)'
    z = quantile.(Normal(), 1 - alpha/2)
    ci_basic = hcat(2 .* object["theta"] .- quants_upp, 2 .* object["theta"] .- quants_low)
    ci_perc = hcat(quants_low, quants_upp)
    ci_norm = hcat(2 .* object["theta"] .- means .- z .* sds,  2 .* object["theta"] .- means .+ z .* sds)
    valid = sum(.!isnan.(coefs), dims = 1)'
    Dict("valid_samples" => valid,
         "means" => means[:, 1],
         "ses" => sds[:, 1],
         "ci_basic" => ci_basic,
         "ci_perc" => ci_perc,
         "ci_norm" => ci_norm)
    
end

## A helper function for simulations
function get_results(data::Data,
                     start::Vector{Float64},
                     br::Bool = false;
                     subsets::Union{Nothing, Vector{Vector{Int64}}} = nothing,
                     plugin::Bool = false,
                     optim_options = Optim.Options())
    pall = nvar(data) + 1
    coefs = fill(NaN, pall) 
    converged = fill(false, 1)
    obje = fill(NaN, 1) 
    gradient = fill(NaN, pall)
    ses = fill(NaN, pall)
    try
        temp_fit = fit(start, data, br, Mstart = false, subsets = subsets, plugin = plugin,
                       optim_options = optim_options)
        coefs[:] = temp_fit.minimizer
        converged[1] = Optim.converged(temp_fit)
        gradient[:] = grad(coefs, data, br, subsets = subsets, plugin = plugin)
        obje[1] = -Optim.minimum(temp_fit)
    catch
    end
    ses[:] = se(coefs, data, subsets = subsets)
    Dict("coefs" => coefs,
         "converged" => converged[1],
         "grad" => gradient,
         "ses" => ses,
         "start" => start,
         "objective" => obje[1])
end

function progress(i::Int64, out_of::Int64, prefix::String, verbose)
    if (i % verbose == 0)
        println(prefix, round(i / out_of  * 100, digits = 4), "%",
                "\t Run: ", i, " out of ", out_of)
    end
end

end
