## Copyright (C) 2022 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!
##
## It is assumed that the user has a Julia and R installation with all
## packages required by the current script and any other script
## included.
##
##
## Code to reproduce the results in Figure 2

library("ggplot2")
library("patchwork")
library("dplyr")
library("colorspace")

images <- dir("images", patter = "bias_", full.names = TRUE)
size <- 1

get_df <- function(results, truth, what = "ml") {
    coefs <- sapply(results, function(x) x[[paste0("coef_", what)]]) |> data.frame() |> stack()
    infinite <- sapply(results, function(x) x[["infinite_ml"]]) |> data.frame() |> stack()
    grad <- sapply(results, function(x) c(x[[paste0("scores_", what)]])) |> data.frame() |> stack()
    names(coefs) <- c("estimate", "sample")
    coefs$grad <- grad[, 1]
    coefs$infinite <- infinite[, 1] != 0 | abs(coefs$estimate) > 30
    coefs$sample <- as.numeric(coefs$sample)
    coefs$parameter <- paste0("beta[", 1:length(betas), "]")
    coefs$truth <- truth
    coefs |> filter(!is.na(estimate))
}

coefs <- NULL
separated <- numeric(length(images))
names(separated) <- images
for (image in images) {
    load(image)
    coef_ml <- get_df(res, betas, what = "ml") |> mutate(method = "ML", n = n_obs)
    coef_br <- get_df(res, betas, what = "br") |> mutate(method = "AS", n = n_obs)
    coef_iRBM <- get_df(res, betas, what = "tr") |> mutate(method = "iRBM", n = n_obs)
    coef_iRBMp <- get_df(res, betas, what = "tr_extra") |> mutate(method = "iRBMp", n = n_obs)
    coef_eRBM <- get_df(res, betas, what = "erbm") |> mutate(method = "eRBM", n = n_obs)
    coefs <- rbind(coefs, rbind(coef_ml, coef_br, coef_iRBM, coef_iRBMp, coef_eRBM))
    separated[image] <- sum(sapply(res, function(x) x$separated))
}

coefs$method <- factor(coefs$method, levels = c("ML", "iRBM", "iRBMp", "eRBM", "AS"), ordered = TRUE)


## Summaries conditional on the ML estimate being finite
summaries <- coefs |>
    group_by(n, method, parameter) |>
    summarize(
        expectation = mean(estimate[!infinite], na.rm = TRUE),
        bias = mean((estimate - truth)[!infinite], na.rm = TRUE),
        rmse = sqrt(mean((estimate - truth)[!infinite]^2, na.rm = TRUE)),
        sd = sd(estimate[!infinite], na.rm = TRUE),
        maxabsgrad = max(abs(grad)),
        rmseOsd = rmse/sd,
        noverlapping  = sum(!infinite),
        mcsd = sd / sqrt(noverlapping),
        pu = mean(estimate < truth, na.rm = TRUE),
        min = min(estimate[!infinite], na.rm = TRUE),
        max = max(estimate[!infinite], na.rm = TRUE),
        prop_valid = noverlapping / n()) |>
    data.frame()

bias <-
    ggplot(summaries) +
    geom_col(aes(x = n, y = abs(bias), fill = method), position = "dodge2") +
    geom_hline(aes(yintercept = 0), colour = "black", lty = 2, alpha = 0.5) +
    facet_wrap(~ parameter, scales = "free_x", nrow = 1, labeller = label_parsed) +
    labs(x = expression(n), y = "|Bias|") +
    theme_minimal() +
    theme(legend.position = "top") +
    scale_fill_grey() +
    scale_x_continuous(breaks = unique(summaries$n), labels = unique(summaries$n))

rmse <-
    ggplot(summaries) +
    geom_col(aes(x = n, y = rmse^2, fill = method), position = "dodge2") +
    facet_wrap(~ parameter, scales = "free_x", nrow = 1, labeller = label_parsed) +
    labs(x = expression(n), y = "RMSE") +
    theme_minimal() +
    theme(legend.position = "bottom") +
    scale_fill_grey()+
    scale_x_continuous(breaks = unique(summaries$n), labels = unique(summaries$n))

noxticks <- theme(axis.ticks.x = element_blank(), axis.text.x = element_blank())
nolegend <- theme(legend.position = "none")
nolabels <- theme(strip.text.x = element_blank())

pdf("probit_bias.pdf", width = 7, height = 3)
(bias + labs(x = NULL) + noxticks + nolegend) /
    (rmse + nolabels + nolegend)
dev.off()




