## Copyright (C) 2022 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!
##
## It is assumed that the user has a Julia and R installation with all
## packages required by the current script and any other script
## included.
##
##
## Code used by the script that reproduces the results in Section S8


p_k <- function(k, trans = "identity") {
    switch(trans,
           "log" = log(k),
           "identity" = k,
           "inverse" = 1/k)
}

k_p <- function(p, trans = "identity") {
    switch(trans,
           "log" = exp(p),
           "identity" = p,
           "inverse" = 1/p)
}

dk_p <- function(p, trans = "identity") {
    switch(trans,
           "log" = exp(p),
           "identity" = 1,
           "inverse" = -1/p^2)
}

d2k_p <- function(p, trans = "identity") {
    switch(trans,
           "log" = exp(p),
           "identity" = 0,
           "inverse" = 2/p^3)
}


loglik_nb <- function(pars, Y, X, trans = "identity", br = FALSE) {
    p <- ncol(X)
    beta <- pars[-(p + 1)]
    eta <- drop(X %*% beta)
    mu <- exp(eta)
    kappa <- k_p(pars[(p + 1)], trans = trans) ## pars[(p + 1)] = log(alpha) when V(Y) = mu + alpha * mu^2
    theta <- 1/kappa ##
    out <- sum(lgamma(theta + Y) - lgamma(theta) + Y * log(mu) + theta * log(theta) - (theta + Y) * log(theta + mu))
    if (br) {
        out <- out + rbm_penalty(pars, Y, X, trans)
    }
    if (is.na(out) | is.infinite(out)) {
        -Inf
    }
    else {
        out
    }
    out
}

score_nb <- function(pars, Y, X, trans = "identity",
                     elementwise = FALSE) {
    p <- ncol(X)
    n <- length(Y)
    beta <- pars[-(p + 1)]
    eta <- drop(X %*% beta)
    mu <- exp(eta)
    kappa <- k_p(pars[(p + 1)], trans = trans)
    lambda <- kappa * mu + 1
    wb <- 1 / lambda
    ## d <- mu
    ## v <- mu + kappa * mu^2
    ## w <- d^2 / v
    ## IK, 26 Dec 2022, Implement this in brglm2 to avoid using s1fun
    inds <- 1:(max(Y) - 1)
    SS <- cumsum(1 / (kappa + 1/inds))
    S1 <- sapply(Y, function(x) if (x < 2) 0 else SS[x - 1])
    ## IK: END
    scores <- cbind(X * wb * (Y - mu),
                    dk_p(p_k(kappa, trans = trans), trans = trans) * (S1 - mu * Y / lambda + (lambda * log(lambda) - lambda + 1) / (kappa^3 * mu + kappa^2)))
    if (elementwise) scores else colSums(scores)
}

## all(sapply(1:n_simu, \(j) all.equal(score_nb(pars0, simu_data[, j], X, "log"),
##                                 numDeriv::grad(loglik_nb, pars0, Y = simu_data[, j], X = X, trans = "log"), check.attributes = FALSE,
##                                 tolerance = 1e-06)))
## all(sapply(1:n_simu, \(j) all.equal(score_nb(pars0, simu_data[, j], X, "inverse"),
##                                     numDeriv::grad(loglik_nb, pars0, Y = simu_data[, j], X = X, trans = "inverse"), check.attributes = FALSE,
##                                     tolerance = 1e-06)))
## all(sapply(1:n_simu, \(j) all.equal(score_nb(pars0, simu_data[, j], X, "identity"),
##                                 numDeriv::grad(loglik_nb, pars0, Y = simu_data[, j], X = X, trans = "identity"), check.attributes = FALSE,
##                                 tolerance = 1e-06)))


j_fun1 <- function(pars, Y, X, trans = "identity") {
    -numDeriv::jacobian(function(pp) score_nb(pp, Y = Y, X = X, trans = trans,
                                              elementwise = FALSE), x = pars)
}

j_fun <- function(pars, Y, X, trans = "identity") {
    p <- ncol(X)
    n <- length(Y)
    beta <- pars[-(p + 1)]
    eta <- drop(X %*% beta)
    mu <- exp(eta)
    kappa <- k_p(pars[(p + 1)], trans = trans)
    lambda <- 1 + kappa * mu
    wb <- 1 / lambda
    wh <- - mu * wb^2
    wbb <- kappa * wh
    e <- Y - mu
    inds <- 1:(max(Y) - 1)
    SS1 <- cumsum(1 / (kappa + 1/inds))
    SS2 <- -cumsum(1 / (kappa + 1/inds)^2)
    S1 <- sapply(Y, function(x) if (x < 2) 0 else SS1[x - 1])
    S2 <- sapply(Y, function(x) if (x < 2) 0 else SS2[x - 1])
    jbb <- crossprod(X * (wb * mu - wbb * e), X)
    phi <- p_k(kappa, trans = trans)
    d1 <- dk_p(phi, trans = trans)
    d2 <- d2k_p(phi, trans = trans)
    jbp <- - d1 * colSums(X * wh * e)
    sp <- sum(S1 - mu * Y / lambda + (lambda * log(lambda) - lambda + 1) / (kappa^3 * mu + kappa^2))
    denom <- kappa^3 * mu + kappa^2
    jpp <- - d2 * sp - d1^2 * sum(S2 + mu^2 * Y / lambda^2 + mu * log(lambda) / denom - (3 * kappa^2 * mu + 2 * kappa) * (lambda * log(lambda) - lambda + 1) / denom^2)
    rbind(cbind(jbb, jbp), c(jbp, jpp))
}

## all(sapply(1:n_simu, \(j) all.equal(j_fun(pars0, simu_data[, j], X, "log"),
##                                     j_fun1(pars0, simu_data[, j], X, "log"),
##                                     check.attributes = FALSE,
##                                     tolerance = 1e-06)))
## all(sapply(1:n_simu, \(j) all.equal(j_fun(pars0, simu_data[, j], X, "identity"),
##                                     j_fun1(pars0, simu_data[, j], X, "identity"),
##                                     check.attributes = FALSE,
##                                     tolerance = 1e-06)))
## all(sapply(1:n_simu, \(j) all.equal(j_fun(pars0, simu_data[, j], X, "inverse"),
##                                     j_fun1(pars0, simu_data[, j], X, "inverse"),
##                                     check.attributes = FALSE,
##                                     tolerance = 1e-06)))

e_fun <- function(pars, Y, X, trans = "identity") {
    crossprod(score_nb(pars, Y = Y, X = X, trans = trans, elementwise = TRUE))
}

rbm_penalty <- function(pars, Y, X, trans = "identity") {
    j_inv <- try(solve(j_fun(pars, Y, X, trans = trans)), silent = TRUE)
    e <- e_fun(pars, Y, X, trans = trans)
    -sum(diag(j_inv %*% e)) / 2
}

bias <- function(pars, Y, X, trans = "identity") {
    A <- numDeriv::grad(rbm_penalty, x = pars, Y = Y, X = X, trans = trans)
    j_inv <- try(solve(j_fun(pars, Y, X, trans = trans)), silent = TRUE)
    - j_inv %*% A
}

## This is in beta, phi parameterization
simulate_nb <- function(pars, X, trans = "identity") {
    n <- nrow(X)
    p <- ncol(X)
    beta <- pars[-(p + 1)]
    eta <- drop(X %*% beta)
    mu <- exp(eta)
    kappa <- k_p(pars[(p + 1)], trans = trans)
    theta <- 1/kappa ##
    rnegbin(n, mu, theta)
}

detect_boundary <- function(pars, Y, X, trans = "identity", kappa_bd = 0.001, se_bd = 100) {
    npar <- ncol(X) + 1
    jinv <- solve(j_fun(pars, Y, X, trans))
    ses <- sqrt(diag(jinv))
    bd <- is.na(ses) | sapply(ses > se_bd, isTRUE)
    if (trans == "identity") {
        bd[npar] <- bd[npar] | pars[npar] < kappa_bd
    }
    if (trans == "inverse") {
        bd[npar] <- bd[npar] | pars[npar] < 1/kappa_bd
    }
    bd
}

has_failed <- function(out, gr_bd = 0.1) {
    (any(is.na(out$estimate)) | isTRUE(any(abs(out$gradient) > gr_bd)) | any(out$boundary))
}

fit_ml <- function(Y, X, start, trans = "identity", kappa_bd = 0.001, se_bd = 100, method = c("L-BFGS-B", "nlminb")) {
    npar <- ncol(X) + 1
    ## out <- try(MASS::glm.nb(Y ~ -1 + X, link = "log", start = start[-(npar - 1)]))
    ## if (inherits(out, "try-error")) {
    lb <- c(rep(-Inf, npar - 1), if (trans == "log") -Inf else 0)
    ub <- rep(Inf, npar)
    out <- try(optimx(start,
                      function(pars) -loglik_nb(pars, Y, X, trans, br = FALSE),
                      function(pars) -score_nb(pars, Y, X, trans, elementwise = FALSE),
                      method = method,
                      lower = lb, upper = ub))
    if (inherits(out, "try-error")) {
        coefs <- rep(NA, npar)
        gr <- rep(NA, npar)
        meth <- NA
    } else {
        oo <- summary(out, order = "value")
        coefs <- unlist(oo[1, seq.int(start)])
        gr <- attr(oo, "details")[, "ngatend"][[1]]
        meth <- rownames(oo)[1]
    }
    ## } else {
    ##         coefs <- c(coef(out), p_k(1/out$theta, trans = trans))
    ##         gr <- score_nb(coefs, Y, X, tr)
    ##         meth <- "glm.nb"
    ##     }
    out <- data.frame(estimate = coefs, gradient = gr, method = "ML")
    attr(out, "omethod") <- meth
    bd <- try(detect_boundary(coefs, Y, X, trans, kappa_bd, se_bd))
    if (inherits(bd, "try-error")) {
        out$boundary <- rep(TRUE, npar)
    } else {
        out$boundary <- bd
    }
    out$estimate[out$boundary] <- NA
    out
}

fit_br <- function(Y, X, start, trans = "identity", kappa_bd = 0.001, se_bd = 100) {
    npar <- ncol(X) + 1
    out <- try(brnb(Y ~ -1 + X, link = "log", transformation = trans, start = start, type = "AS_mean"))
    if (inherits(out, "try-error")) {
        coefs <- gr <- rep(NA, npar)
        meth <- NA
    } else {
        coefs <- coef(out, model = "full")
        gr <- out$grad
        meth <- "brnb"
    }
    out <- data.frame(estimate = coefs, gradient = gr, method = "AS")
    attr(out, "omethod") <- meth
    bd <- try(detect_boundary(coefs, Y, X, trans, kappa_bd, se_bd))
    if (inherits(bd, "try-error")) {
        out$boundary <- rep(TRUE, npar)
    } else {
        out$boundary <- bd
    }
    out$estimate[out$boundary] <- NA
    out
}

fit_irbm <- function(Y, X, start, trans = "identity", kappa_bd = 0.001, se_bd = 100, method = c("L-BFGS-B", "nlminb")) {
    npar <- ncol(X) + 1
    lb <- c(rep(-Inf, npar - 1), if (trans == "log") -Inf else 0)
    ub <- rep(Inf, npar)
    out <- try(optimx(start,
                      function(pars) -loglik_nb(pars, Y, X, trans, br = TRUE),
                      method = method,
                      lower = lb, upper = ub))
    if (inherits(out, "try-error")) {
        coefs <- gr <- rep(NA, npar)
        meth <- NA
    } else {
        oo <- summary(out, order = "value")
        coefs <- unlist(oo[1, seq.int(start)])
        gr <- attr(oo, "details")[, "ngatend"][[1]] ## Only when 2 or more methods are supplied to optimx
        meth <- rownames(oo)[1]
    }
    out <- data.frame(estimate = coefs, gradient = gr, method = "iRBM")
    attr(out, "omethod") <- meth
    bd <- try(detect_boundary(coefs, Y, X, trans, kappa_bd, se_bd))
    if (inherits(bd, "try-error")) {
        out$boundary <- rep(TRUE, npar)
    } else {
        out$boundary <- bd
    }
    out$estimate[out$boundary] <- NA
    out
}

fit_erbm <- function(Y, X, ml_estimates, trans = "identity", kappa_bd = 0.001, se_bd = 100) {
    npar <- ncol(X) + 1
    coefs <- try(ml_estimates - bias(ml_estimates, Y, X, trans))
    if (inherits(coefs, "try-error")) {
        coefs <- gr <- rep(NA, npar)
    } else {
        gr <- score_nb(ml_estimates, Y, X, tr) ## loglik gradient at ML estimates
    }
    out <- data.frame(estimate = coefs, gradient = gr, method = "eRBM")
    bd <- try(detect_boundary(coefs, Y, X, trans, kappa_bd, se_bd))
    if (inherits(bd, "try-error")) {
        out$boundary <- rep(TRUE, npar)
    } else {
        out$boundary <- bd
    }
    out$estimate[out$boundary] <- NA
    out
}
