## Copyright (C) 2022 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!

## Code to reproduce the simulation experiment in Example 3.1

## Data
library("mvtnorm")
library("data.table")
library("copula")
library("parallel")

## path to save intermediate results
path <- "./images/"

## number of cores
n_cores <- 10

## number of rows of base X matrix
base_n <- 5

## number of base simulations
base_n_simu <- 250 * 2^16 ## 2000

## n_obs and n_simu
multiplier <- 2^(0:6)
n_obs <- base_n * multiplier
n_simu <- base_n_simu / n_obs

## Bootstrap size
Rb <- 500

## Parameter setting
setting <- list(sigma = matrix(c(1, 0.5, 0.5, 1), ncol = 2),
                mu = c(2, 10))

sigma <- setting$sigma
mu <- setting$mu
theta <- mu[2]/mu[1]

set.seed(123)
seeds <- round(runif(length(n_obs), 0, 10^6))

simu_fun <- function(nsimu, mu, sigma) {
    out <- rCopula(n = nsimu, normalCopula(sigma[1,2], 2))
    t(apply(out, 1, function(x) c(qexp(x[1], 1/mu[1]), qnorm(x[2], mu[2], sqrt(sigma[2, 2])))))
}

boot <- function(D, R = 100) {
    x <- D[, 1]
    y <- D[, 2]
    N <- nrow(D)
    finds <- function(x) {
        v <- structure(rep(0, N), names = 1:N)
        pr <- table(x) / N
        v[names(pr)] <- pr
        v
    }
    inds <- replicate(R, sample(1:N, N, replace = TRUE))
    w <- rowMeans(apply(inds, 2, finds))
    mle <- sum(y) / sum(x)
    theta_star <- mean(apply(inds, 2, function(ind) {
        sum(y[ind])/sum(x[ind])
    }))
    Tr <- sum(y * w) / sum(x * w)
    mle - theta_star + structure(c(Tr, mle), names = c("better", "naive"))
}

estimate_ratio <- function(D, truth, R = 100) {
    sx <- sum(D[, 1])
    sy <- sum(D[, 2])
    sxy <- sum(D[, 1] * D[, 2])
    sxx <- sum(D[, 1] * D[, 1])
    syy <- sum(D[, 2] * D[, 2])
    n <- nrow(D)
    mle <- sy / sx
    br <- (sy * sx + sxy) / (sx^2 + sxx)
    bc <- mle * (1 - sxx / sx^2) + sxy/ sx^2
    mle_i <- sapply(1:n, function(i) { m <- colSums(D[-i, ]); m[2]/m[1] })
    jack <- n * mle - (n - 1)/n * sum(mle_i)
    boot_ests <- boot(D, R = R)
    e <- function(theta) (theta^2 * sxx + syy - 2 * theta * sxy)
    v <- function(theta) e(theta)/sx^2
    sbr <- function(theta) sy - theta * sx + sxy/sx - theta * sxx/sx
    s <- function(theta) sy - theta * sx
    br_var <- v(br)
    bc_var <- v(bc)
    mle_var <- v(mle)
    jack_var <- sum((mle_i - mean(mle_i))^2) * (n - 1) / n
    br_pvalue_e <- 1 - pchisq((br - truth)^2/br_var, 1)
    bc_pvalue_e <- 1 - pchisq((bc - truth)^2/bc_var, 1)
    mle_pvalue_e <- 1 - pchisq((mle - truth)^2/mle_var, 1)
    mle_pvalue_s <- 1 - pchisq(s(truth)^2 / e(truth), 1)
    br_pvalue_s <- 1 - pchisq(sbr(truth)^2 / e(truth), 1)
    k <- attr(D, "sample")
    out <- data.frame(parameter = c("theta", "theta", "theta", "theta", "theta", "theta"),
                      method = c("ML", "BR", "Jackknife", "BC", "Bootstrap[better]", "Bootstrap[naive]"),
                      estimate = c(mle, br, jack, bc, boot_ests),
                      evar = c(mle_var, br_var, jack_var, bc_var, NA, NA),
                      pvalue_e = c(mle_pvalue_e, br_pvalue_e, NA, bc_pvalue_e, NA, NA),
                      pvalue_s = c(mle_pvalue_s, br_pvalue_s, NA, NA, NA, NA))
    out
}

fit_fun <- function(k) {
    DD <- simu_fun(n_obs[j], mu, sigma)
    out <- estimate_ratio(DD, truth = mu[2]/mu[1], R = Rb)
    out$sample <- k
    if (k %% 1000 == 0) cat("n: ", nrow(DD), "size:", n_simu[j], "|",  "sample:", k, "Done!\n")
    out
}


RNGkind("L'Ecuyer-CMRG")
for (j in 1:length(n_simu)) {
    ## Simulate data
    set.seed(seeds[j])
    results <- mclapply(1:n_simu[j], fit_fun, mc.cores = n_cores, mc.set.seed = TRUE)
    ## Save
    results <- rbindlist(results)
    results$true_value <- theta
    attr(results, "N") <- n_simu[j]
    attr(results, "n") <- n_obs[j]
    attr(results, "seed") <- seeds[j]
    save(list = "results",
         file = paste0(path, "/ratio-setting-copula", "-n-", j, ".rda"))
}

