## Copyright (C) 2022 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!
##
## It is assumed that the user has a Julia and R installation with all
## packages required by the current script and any other script
## included.
##
##
## Code to reproduce the results in Table 2

library("parallel")
library("dplyr")
library("ggplot2")

## Path to read images from
path <- "./images/"

files <- dir(path, full.names = TRUE)
files <- files[grep("ratio-setting-copula", files)]

all_results <- NULL
for (f in files) {
    load(f)
    cres <- results
    cres$n <- attr(cres, "n")
    cres$N <- attr(cres, "N")
    all_results <- rbind(all_results, cres)
}

M <- Inf
summaries <- NULL
for (j in seq_along(M)) {
    summaries[[j]] <- all_results %>%
        group_by(n, parameter, method) %>%
        summarize(min_estimate = min(estimate),
                  max_estimate = max(estimate),
                  bias = mean((estimate - true_value)),
                  mad = mean(abs(estimate - true_value)),
                  pu = mean(estimate < true_value),
                  variance = var(estimate),
                  est_variance = mean(evar),
                  mse = bias^2 + variance,
                  typeI_e = mean(pvalue_e < 0.05),
                  typeI_s = mean(pvalue_s < 0.05),
                  N = n(),
                  mc_error = sd(estimate)/sqrt(N),
                  scaled_bias = bias * unique(n),
                  scaled_bias_mc_error = mc_error * unique(n),
                  M = M[j])
}

summaries <- summaries[[1]]



## Output for latex

what <- c("bias", "mse", "mad", "pu")

tab <- NULL
summaries_wide <- NULL
for (w in what) {
    summs <- as.data.frame(summaries[, c("n", "method", w)])
    summs <- reshape(summs, v.names = w, direction = "wide", idvar = "method", timevar ="n")
    summaries_wide[[w]] <- summs
    summs <- data.frame(summary = w, summs, stringsAsFactors = FALSE)
    nams <- names(summs)
    names(summs) <- gsub(paste0(w, "."), "n=", nams)
    summs <- summs[c(6, 5, 3, 2, 4, 1), ]
    rownames(summs) <- NULL
    tab <- rbind(tab, summs)
}
names(tab)[which(names(tab) == "method")] <- "estimator"
names(tab)[which(names(tab) == "summary")] <- ""

memisc:::toLatex.data.frame(tab, useDcolumn = FALSE, digits = 2)


## Slopes

df <- reshape(summaries_wide$bias, direction = "long")
m_ML <- lm(I(log(abs(bias))) ~ log(n), data = df, subset = method == "ML" & n > 4 & n < 321)
m_BR <- lm(I(log(abs(bias))) ~ log(n), data = df, subset = method == "BR" & n > 4 & n < 321)
m_BC <- lm(I(log(abs(bias))) ~ log(n), data = df, subset = method == "BC" & n > 4 & n < 321)
m_Ja <- lm(I(log(abs(bias))) ~ log(n), data = df, subset = method == "Jackknife"  & n < 321)
m_BB <- lm(I(log(abs(bias))) ~ log(n), data = df, subset = method == "Bootstrap[better]"  & n < 321)

round(sapply(list(m_ML, m_Ja, m_BB, m_BC, m_BR), coef)[2,], 3)
